# noqa: D100
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

from slack_bolt import App
from slack_sdk.errors import SlackApiError

from django_slack_tools.slack_messages.models import SlackMessage, SlackMessageRecipient, SlackMessagingPolicy
from django_slack_tools.slack_messages.request import MessageHeader, MessageRequest

from .base import BaseMiddleware

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.messenger import Messenger
    from django_slack_tools.slack_messages.response import MessageResponse

logger = logging.getLogger(__name__)


class DjangoDatabasePersister(BaseMiddleware):
    """Persist message history to database. If request is `None`, will do nothing."""

    def __init__(self, *, slack_app: App | None = None, get_permalink: bool = False) -> None:
        """Initialize the middleware.

        Args:
            slack_app: Slack app instance to use for certain tasks, such as getting permalinks.
            get_permalink: If `True`, will try to get the permalink of the message.
        """
        if get_permalink and not isinstance(slack_app, App):
            msg = "`slack_app` must be an instance of `App` if `get_permalink` is set `True`."
            raise ValueError(msg)

        self.slack_app = slack_app
        self.get_permalink = get_permalink

    def process_response(self, response: MessageResponse) -> MessageResponse | None:  # noqa: D102
        request = response.request
        if request is None:
            logger.warning("No request found in response, skipping persister.")
            return response

        logger.debug("Getting permalink for message: %s", response)
        if self.get_permalink:  # noqa: SIM108
            permalink = self._get_permalink(channel=request.channel, ts=response.ts)
        else:
            permalink = ""

        logger.debug("Persisting message history to database: %s", response)
        try:
            history = SlackMessage(
                id=request.id_,
                channel=request.channel,
                header=request.header.model_dump(),
                body=request.body.model_dump() if request.body else {},
                ok=response.ok,
                permalink=permalink,
                ts=response.ts,
                parent_ts=response.parent_ts or "",
                request=request.model_dump(),
                response=response.model_dump(exclude={"request"}),
                exception=response.error or "",
            )
            history.save()
        except Exception:
            logger.exception("Error while saving message history: %s", response)

        return response

    def _get_permalink(self, *, channel: str, ts: str | None) -> str:
        """Get permalink of the message. It returns empty string on error."""
        if not self.slack_app:
            logger.warning("Slack app not provided, cannot get permalink.")
            return ""

        if not ts:
            logger.warning("No message ts provided, cannot get permalink.")
            return ""

        try:
            response = self.slack_app.client.chat_getPermalink(channel=channel, message_ts=ts)
            return response.get("permalink", default="")
        except SlackApiError as err:
            logger.debug("Error while getting permalink: %s", exc_info=err)
            return ""


OnPolicyNotExists = Literal["create", "default", "error"]


class DjangoDatabasePolicyHandler(BaseMiddleware):
    """Middleware to handle Slack messaging policies stored in the database.

    Be cautious when using this middleware because it includes functionality to distribute messages to multiple recipients,
    which could lead to unwanted infinite loop or recursion if used improperly.

    This middleware contains a secondary protection against infinite loops by injecting a context key to the message context.
    If the key is found in the context, the middleware will stop the message from being sent. So be careful when modifying the context.
    """  # noqa: E501

    _RECURSION_DETECTION_CONTEXT_KEY = "__final__"
    """Recursion detection key injected to message context for fanned-out messages to provide secondary protection against infinite loops."""  # noqa: E501

    def __init__(
        self,
        *,
        messenger: Messenger | str,
        on_policy_not_exists: OnPolicyNotExists = "error",
    ) -> None:
        """Initialize the middleware.

        This middleware will load the policy from the database and send the message to all recipients.

        Args:
            messenger: Messenger instance or name to use for sending messages.
                The messenger instance should be different from the one used in the policy handler,
                because this middleware cannot properly handle fanned-out messages modified by this middleware.
                Also, there are chances of infinite loops if the same messenger is used.
            on_policy_not_exists: Action to take when policy is not found.
        """
        if on_policy_not_exists not in ("create", "default", "error"):
            msg = f'Unknown value for `on_policy_not_exists`: "{on_policy_not_exists}"'
            raise ValueError(msg)

        self._messenger = messenger
        self.on_policy_not_exists = on_policy_not_exists

    # * It's not desirable to put import in the method,
    # * but it's the only way to avoid circular imports for now (what's the fix?)
    @property
    def messenger(self) -> Messenger:
        """Get the messenger instance. If it's a string, will get the messenger from the app settings."""
        if isinstance(self._messenger, str):
            from django_slack_tools.app_settings import get_messenger

            self._messenger = get_messenger(self._messenger)

        return self._messenger

    def process_request(self, request: MessageRequest) -> MessageRequest | None:  # noqa: D102
        # TODO(lasuillard): Hacky way to stop the request, need to find a better way
        #                   Some extra field (request.meta) could be added to share control context
        if request.context.get(self._RECURSION_DETECTION_CONTEXT_KEY, False):
            return request

        code = request.channel
        policy = self._get_policy(code=code)
        if not policy.enabled:
            logger.debug("Policy %s is disabled, skipping further messaging", policy)
            return None

        requests: list[MessageRequest] = []
        for recipient in policy.recipients.all():
            default_context = self._get_default_context(recipient)
            context = {
                **default_context,
                **request.context,
                self._RECURSION_DETECTION_CONTEXT_KEY: True,
            }
            header = MessageHeader.model_validate(
                {
                    **policy.header_defaults,
                    **request.header.model_dump(),
                },
            )
            req = MessageRequest(channel=recipient.channel, template_key=policy.code, context=context, header=header)
            requests.append(req)

        # TODO(lasuillard): How to provide users the access the newly created messages?
        #                   currently, it's possible with persisters but it would require some additional work
        # TODO(lasuillard): Can `sys.setrecursionlimit` be used to prevent spamming if recursion occurs?
        for req in requests:
            self.messenger.send_request(req)

        # Stop current request
        return None

    def _get_policy(self, *, code: str) -> SlackMessagingPolicy:
        """Get the policy for the given code."""
        try:
            policy = SlackMessagingPolicy.objects.get(code=code)
        except SlackMessagingPolicy.DoesNotExist:
            if self.on_policy_not_exists == "create":
                logger.warning("No policy found for template key, creating one: %s", code)
                policy = self._create_policy(code=code)
            elif self.on_policy_not_exists == "default":
                policy = SlackMessagingPolicy.objects.get(code="DEFAULT")
            elif self.on_policy_not_exists == "error":
                raise
            else:
                msg = f'Unknown value for `on_policy_not_exists`: "{self.on_policy_not_exists}"'
                raise ValueError(msg) from None

        return policy

    def _create_policy(self, *, code: str) -> SlackMessagingPolicy:
        """Create a policy with the given code.

        Policy created is disabled by default, thus no message will be sent.
        To modify the default policy creation behavior, simply override this method.
        """
        policy = SlackMessagingPolicy.objects.create(
            code=code,
            enabled=False,
            template_type=SlackMessagingPolicy.TemplateType.UNKNOWN,
        )
        default_recipients = SlackMessageRecipient.objects.filter(alias="DEFAULT")
        policy.recipients.set(default_recipients)
        return policy

    def _get_default_context(self, recipient: SlackMessageRecipient) -> dict:
        """Create default context for the recipient."""
        mentions = [mention.mention for mention in recipient.mentions.all()]
        return {
            "mentions": mentions,
        }



"""Application settings."""
# flake8: noqa: UP007

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, TypedDict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from slack_bolt import App
from typing_extensions import NotRequired, Self

from django_slack_tools.utils.import_helper import LazyInitSpec, lazy_init

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.messenger import Messenger

logger = getLogger(__name__)


class SettingsDict(TypedDict):
    """Dictionary input for app settings."""

    slack_app: str
    messengers: NotRequired[dict[str, LazyInitSpec]]


class AppSettings:
    """Application settings."""

    slack_app: App
    messengers: dict[str, Messenger]

    def __init__(self, slack_app: App, messengers: dict[str, Messenger]) -> None:  # noqa: D107
        if not isinstance(slack_app, App):
            msg = f"Expected {App!s} instance, got {type(slack_app)}"
            raise TypeError(msg)

        self.slack_app = slack_app
        self.messengers = messengers

    @classmethod
    def from_dict(cls, settings_dict: SettingsDict) -> Self:
        """Initialize settings from a dictionary."""
        try:
            slack_app = import_string(settings_dict["slack_app"])
            messengers = {
                name: cls._create_messenger(spec) for name, spec in settings_dict.get("messengers", {}).items()
            }
            return cls(
                slack_app=slack_app,
                messengers=messengers,
            )
        except Exception as err:
            msg = f"Couldn't initialize app settings: {err!s}"
            raise ImproperlyConfigured(msg) from err

    @classmethod
    def _create_messenger(cls, spec: LazyInitSpec) -> Messenger:
        class_ = import_string(spec["class"])
        args, kwargs = spec.get("args", ()), spec.get("kwargs", {})

        # Lazy-init all the things
        kwargs = kwargs.copy()
        kwargs["template_loaders"] = [lazy_init(tl) for tl in kwargs["template_loaders"]]
        kwargs["middlewares"] = [lazy_init(tl) for tl in kwargs["middlewares"]]
        kwargs["messaging_backend"] = lazy_init(kwargs["messaging_backend"])

        # Create the messenger
        return class_(*args, **kwargs)  # type: ignore[no-any-return]


def get_settings_from_django(settings_key: str = "DJANGO_SLACK_TOOLS") -> AppSettings:
    """Get application settings."""
    django_settings: SettingsDict = getattr(settings, settings_key)
    return AppSettings.from_dict(django_settings or {})


app_settings = get_settings_from_django()


def get_messenger(name: str | None = None) -> Messenger:
    """Get a messenger instance by name."""
    name = name or "default"
    return app_settings.messengers[name]



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest
from slack_bolt import App

from django_slack_tools.app_settings import AppSettings

if TYPE_CHECKING:
    from collections.abc import Iterator

    from django_slack_tools.app_settings import SettingsDict


@pytest.fixture(scope="session")
def slack_app() -> App:
    """Dummy Slack app fixture. It won't work."""
    return App(
        token="stupid-sandwich",  # noqa: S106
        signing_secret="peanut-butter",  # noqa: S106
        token_verification_enabled=False,
    )


@pytest.fixture
def mock_slack_client() -> Iterator[mock.Mock]:
    """Mock `slack_bolt.App.client`."""
    with mock.patch("slack_bolt.App.client") as m:
        yield m


@pytest.fixture(scope="session")
def app_settings() -> SettingsDict:
    """Default app settings fixture. Each test suite should override it for practical test cases."""
    return {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [],
                    "middlewares": [],
                    "messaging_backend": "django_slack_tools.slack_messages.backends.DummyBackend",
                },
            },
        },
    }


@pytest.fixture
def override_app_settings(app_settings: SettingsDict) -> Iterator[None]:
    """Override app settings with settings dict returned by `app_settings` fixture."""
    override = AppSettings.from_dict(app_settings)
    with mock.patch("django_slack_tools.app_settings.app_settings", override):
        yield



from __future__ import annotations

import sys
from contextlib import contextmanager
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from django_slack_tools.slack_messages.backends import DummyBackend
from django_slack_tools.slack_messages.messenger import Messenger
from django_slack_tools.slack_messages.middlewares import (
    BaseMiddleware,
    DjangoDatabasePersister,
    DjangoDatabasePolicyHandler,
)
from django_slack_tools.slack_messages.models import SlackMessage
from django_slack_tools.slack_messages.models.messaging_policy import SlackMessagingPolicy
from django_slack_tools.slack_messages.request import MessageRequest
from django_slack_tools.slack_messages.response import MessageResponse
from tests._factories import SlackApiErrorFactory
from tests._helpers import AnyRegex
from tests.slack_messages._factories import (
    MessageRequestFactory,
    MessageResponseFactory,
    SlackGetPermalinkResponseFactory,
)
from tests.slack_messages._helpers import MockBackend, MockTemplateLoader
from tests.slack_messages.models._factories import (
    SlackMentionFactory,
    SlackMessageRecipientFactory,
    SlackMessagingPolicyFactory,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from slack_bolt import App

    from django_slack_tools.app_settings import SettingsDict


pytestmark = [
    pytest.mark.usefixtures("override_app_settings"),
    pytest.mark.django_db,
]


@pytest.fixture(scope="session")
def app_settings() -> SettingsDict:
    return {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "test-django-middleware": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [
                        {
                            "class": "django_slack_tools.slack_messages.template_loaders.DjangoPolicyTemplateLoader",
                            "kwargs": {},
                        },
                    ],
                    "middlewares": [
                        {
                            "class": "django_slack_tools.slack_messages.middlewares.DjangoDatabasePersister",
                            "kwargs": {},
                        },
                    ],
                    "messaging_backend": {
                        "class": "django_slack_tools.slack_messages.backends.DummyBackend",
                        "kwargs": {},
                    },
                },
            },
        },
    }


class TestDjangoDatabasePersister:
    def test_instance_creation(self, slack_app: App) -> None:
        """Test various instance creation scenarios."""
        DjangoDatabasePersister(slack_app=None, get_permalink=False)
        with pytest.raises(
            ValueError,
            match="`slack_app` must be an instance of `App` if `get_permalink` is set `True`.",
        ):
            DjangoDatabasePersister(slack_app=None, get_permalink=True)

        DjangoDatabasePersister(slack_app=slack_app, get_permalink=True)
        DjangoDatabasePersister(slack_app=slack_app, get_permalink=False)

    def test_process_response(self, slack_app: App, mock_slack_client: mock.Mock) -> None:
        """Test processing the response."""
        permalink = "https://example.com/permalink"
        mock_slack_client.chat_getPermalink.return_value = SlackGetPermalinkResponseFactory(permalink=permalink)
        persister = DjangoDatabasePersister(slack_app=slack_app, get_permalink=True)

        response = MessageResponseFactory.create()
        result = persister.process_response(response)
        saved_message = SlackMessage.objects.get(id=response.request.id_)

        assert result is response
        assert saved_message.policy is None
        assert saved_message.channel == response.request.channel
        assert saved_message.header == {
            "mrkdwn": None,
            "parse": None,
            "reply_broadcast": None,
            "thread_ts": None,
            "unfurl_links": None,
            "unfurl_media": None,
        }
        assert saved_message.body == {}
        assert saved_message.ok == response.ok
        assert saved_message.permalink == permalink
        assert saved_message.ts == response.ts
        assert saved_message.parent_ts == ""
        assert MessageRequest.model_validate(saved_message.request)
        assert MessageResponse.model_validate(saved_message.response)
        assert saved_message.exception == ""

    def test_process_response_fail_save_db(self) -> None:
        """Test failing to save to the database. It should log the error, but not raise it."""
        persister = DjangoDatabasePersister()
        with mock.patch("django_slack_tools.slack_messages.models.SlackMessage.save") as mock_save:
            mock_save.side_effect = Exception("Some error occurred")
            persister.process_response(MessageResponseFactory.create())

    def test_process_response_skip_persist_if_request_none(self) -> None:
        """Test skipping persisting if `response.request` is `None`."""
        persister = DjangoDatabasePersister()

        response = MessageResponseFactory.create(request=None)
        result = persister.process_response(response)

        assert result is response
        assert not SlackMessage.objects.exists()

    def test_get_permalink(self, slack_app: App, mock_slack_client: mock.Mock) -> None:
        """Test getting permalink for the message."""
        expect = "https://example.com/permalink"
        mock_slack_client.chat_getPermalink.return_value = SlackGetPermalinkResponseFactory(permalink=expect)
        persister = DjangoDatabasePersister(slack_app=slack_app, get_permalink=True)
        permalink = persister._get_permalink(channel="test-channel", ts="some-ts")
        assert permalink == expect

    def test_get_permalink_error_api_call_fails(self, slack_app: App, mock_slack_client: mock.Mock) -> None:
        """Test getting permalink error when API call fails."""
        mock_slack_client.chat_getPermalink.side_effect = SlackApiErrorFactory()
        persister = DjangoDatabasePersister(slack_app=slack_app, get_permalink=True)
        permalink = persister._get_permalink(channel="test-channel", ts="some-ts")
        assert permalink == ""

    def test_get_permalink_error_no_slack_app(self, slack_app: App) -> None:
        """Test getting permalink error when slack_app is None."""
        persister = DjangoDatabasePersister(slack_app=slack_app, get_permalink=True)
        persister.slack_app = None  # Bypass the validation in the constructor
        permalink = persister._get_permalink(channel="test-channel", ts="some-ts")
        assert permalink == ""

    def test_get_permalink_error_no_ts(self, slack_app: App) -> None:
        """Test getting permalink error when ts is empty."""
        persister = DjangoDatabasePersister(slack_app=slack_app, get_permalink=True)
        assert persister._get_permalink(channel="test-channel", ts="") == ""


class TestDjangoDatabasePolicyHandler:
    def test_instance_creation(self) -> None:
        """Test various instance creation scenarios."""
        messenger = Messenger(template_loaders=[], middlewares=[], messaging_backend=DummyBackend())

        middleware = DjangoDatabasePolicyHandler(messenger=messenger)
        assert middleware.messenger == messenger

        with pytest.raises(ValueError, match='Unknown value for `on_policy_not_exists`: "unknown-behavior"'):
            middleware = DjangoDatabasePolicyHandler(messenger=messenger, on_policy_not_exists="unknown-behavior")  # type: ignore[arg-type]

        middleware = DjangoDatabasePolicyHandler(messenger="test-django-middleware")
        assert isinstance(middleware.messenger, Messenger)

        with pytest.raises(ValueError, match='Unknown value for `on_policy_not_exists`: "unknown-behavior"'):
            middleware = DjangoDatabasePolicyHandler(
                messenger="test-django-middleware",
                on_policy_not_exists="unknown-behavior",  # type: ignore[arg-type]
            )

    def test_process_request(self) -> None:
        """Test processing the request."""
        # Arrange
        middleware = DjangoDatabasePolicyHandler(messenger="test-django-middleware")
        recipients = [SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(3)) for _ in range(3)]
        policy = SlackMessagingPolicyFactory(
            recipients=recipients,
            template_type=SlackMessagingPolicy.TemplateType.DJANGO_INLINE,
            template="""
<root>
    <block type="section">
        <text type="mrkdwn">
            {{ greet }}, {{ mentions | join:", " }}!
        </text>
    </block>
</root>
            """.strip(),
        )

        # Act
        # Original request will be stopped, and 3 new requests will be created
        response = middleware.process_request(
            MessageRequestFactory.create(
                channel=policy.code,
                context={"greet": "Nice to meet you"},
            ),
        )

        # Assert
        assert response is None
        assert SlackMessage.objects.count() == 3
        assert list(
            SlackMessage.objects.all().values("policy", "channel", "header", "body", "ok"),
        ) == [
            {
                "policy": None,
                "channel": mock.ANY,
                "header": {
                    "mrkdwn": None,
                    "parse": None,
                    "reply_broadcast": None,
                    "thread_ts": None,
                    "unfurl_links": None,
                    "unfurl_media": None,
                },
                "body": {
                    "attachments": None,
                    "blocks": [
                        {
                            "type": "section",
                            "text": {"type": "mrkdwn", "text": AnyRegex(r"Nice to meet you, .*, .*, .*!")},
                        },
                    ],
                    "text": None,
                    "icon_emoji": None,
                    "icon_url": None,
                    "metadata": None,
                    "username": None,
                },
                "ok": True,
            },
            {
                "policy": None,
                "channel": mock.ANY,
                "header": {
                    "mrkdwn": None,
                    "parse": None,
                    "reply_broadcast": None,
                    "thread_ts": None,
                    "unfurl_links": None,
                    "unfurl_media": None,
                },
                "body": {
                    "attachments": None,
                    "blocks": [
                        {
                            "type": "section",
                            "text": {"type": "mrkdwn", "text": AnyRegex(r"Nice to meet you, .*, .*, .*!")},
                        },
                    ],
                    "text": None,
                    "icon_emoji": None,
                    "icon_url": None,
                    "metadata": None,
                    "username": None,
                },
                "ok": True,
            },
            {
                "policy": None,
                "channel": mock.ANY,
                "header": {
                    "mrkdwn": None,
                    "parse": None,
                    "reply_broadcast": None,
                    "thread_ts": None,
                    "unfurl_links": None,
                    "unfurl_media": None,
                },
                "body": {
                    "attachments": None,
                    "blocks": [
                        {
                            "type": "section",
                            "text": {"type": "mrkdwn", "text": AnyRegex(r"Nice to meet you, .*, .*, .*!")},
                        },
                    ],
                    "text": None,
                    "icon_emoji": None,
                    "icon_url": None,
                    "metadata": None,
                    "username": None,
                },
                "ok": True,
            },
        ]

    def test_process_request_recursion_detection(self) -> None:
        """Test recursion detection mechanism. Fanned-out requests should contain special context key for detection."""
        # Arrange
        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[],
            messaging_backend=DummyBackend(),
        )
        messenger.middlewares = [
            DjangoDatabasePolicyHandler(messenger=messenger),
            DjangoDatabasePersister(),
        ]
        policy = SlackMessagingPolicyFactory.create(
            code="test-channel",
            recipients=SlackMessageRecipientFactory.create_batch(size=3, channel="test-channel"),
        )

        # Act
        messenger.send(policy.code, context={"name": "Daniel"})

        # Assert
        assert SlackMessage.objects.all().count() == 3

    @pytest.mark.skip(reason="Can't run this test because pytest session crashes. Is there a way to test this?")
    def test_process_request_force_infinite_recursion(self) -> None:
        """Demonstrate what happens if detection key is corrupted."""
        # Arrange

        # This middleware will corrupt the detection key
        class BadMiddlewareCorruptDetectionKey(BaseMiddleware):
            def process_request(self, request: MessageRequest) -> MessageRequest | None:
                request.context[DjangoDatabasePolicyHandler._RECURSION_DETECTION_CONTEXT_KEY] = "corrupted"
                return request

        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[],
            messaging_backend=DummyBackend(),
        )
        messenger.middlewares = [
            BadMiddlewareCorruptDetectionKey(),
            DjangoDatabasePolicyHandler(messenger=messenger),
            DjangoDatabasePersister(),
        ]
        policy = SlackMessagingPolicyFactory.create(
            code="test-channel",
            recipients=SlackMessageRecipientFactory.create_batch(size=3, channel="test-channel"),
        )

        # Act
        with pytest.raises(RecursionError), recursion_limit(40):
            messenger.send(policy.code, context={"name": "Daniel"})

        # Assert
        assert SlackMessage.objects.all().count() == 3

    def test_process_request_policy_not_exists(self) -> None:
        """If policy does not exists and `on_policy_not_exists` is `"error"`, it should throw an error."""
        # Arrange
        messenger = Messenger(template_loaders=[], middlewares=[], messaging_backend=DummyBackend())
        middleware = DjangoDatabasePolicyHandler(messenger=messenger)

        # Act & Assert
        with pytest.raises(SlackMessagingPolicy.DoesNotExist):
            middleware.process_request(MessageRequestFactory.create(channel="nonexistent-policy-code"))

    def test_process_request_policy_on_policy_not_exists_create(self) -> None:
        """If policy does not exists but `on_policy_not_exists` is `"create"`, it should create the policy."""
        # Arrange
        messenger = Messenger(template_loaders=[], middlewares=[], messaging_backend=DummyBackend())
        middleware = DjangoDatabasePolicyHandler(messenger=messenger, on_policy_not_exists="create")

        # Act
        request = middleware.process_request(
            MessageRequestFactory.create(channel="nonexistent-policy-code", context={}),
        )

        # Assert
        assert request is None

        created_policy = SlackMessagingPolicy.objects.get(code="nonexistent-policy-code")
        assert created_policy.enabled is False
        assert list(created_policy.recipients.values_list("alias", flat=True)) == ["DEFAULT"]
        assert created_policy.header_defaults == {}
        assert created_policy.template_type == SlackMessagingPolicy.TemplateType.UNKNOWN
        assert created_policy.template is None

    def test_process_request_policy_on_policy_not_exists_use_default(self) -> None:
        """If policy does not exists but `on_policy_not_exists` is `"default"`, it should use the default policy."""
        # Arrange
        messenger = Messenger(template_loaders=[], middlewares=[], messaging_backend=DummyBackend())
        middleware = DjangoDatabasePolicyHandler(messenger=messenger, on_policy_not_exists="default")

        # Act
        request = middleware.process_request(
            MessageRequestFactory.create(channel="nonexistent-policy-code", context={}),
        )

        # Assert
        assert request is None
        assert list(SlackMessagingPolicy.objects.all().values_list("code", flat=True)) == ["DEFAULT"]

    def test_process_request_policy_exists_but_disabled(self) -> None:
        """if the policy exists but is disabled, it shouldn't send any message -- simply ignored."""
        # Arrange
        messenger = Messenger(
            template_loaders=[],
            middlewares=[DjangoDatabasePersister()],
            messaging_backend=MockBackend(should_error=True),
        )
        middleware = DjangoDatabasePolicyHandler(messenger=messenger)
        policy = SlackMessagingPolicyFactory(enabled=False)

        # Act
        # Policy is disabled, so no message to send -- shouldn't raise any error.
        request = middleware.process_request(MessageRequestFactory.create(channel=policy.code))

        # Assert
        assert request is None

    def test_process_request_policy_bad_value_for_on_policy_not_exists(self) -> None:
        """If `on_policy_not_exists` is set to a bad value, it should raise an error."""
        # Arrange
        messenger = Messenger(template_loaders=[], middlewares=[], messaging_backend=DummyBackend())
        middleware = DjangoDatabasePolicyHandler(messenger=messenger)
        middleware.on_policy_not_exists = "unknown-behavior"  # type: ignore[assignment]

        # Act & Assert
        with pytest.raises(ValueError, match='Unknown value for `on_policy_not_exists`: "unknown-behavior"'):
            middleware.process_request(MessageRequestFactory.create())


@contextmanager
def recursion_limit(new_limit: int) -> Iterator[None]:
    """Set the recursion limit to a low value."""
    old_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(new_limit)
    yield
    sys.setrecursionlimit(old_limit)



# noqa: D100
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django.template import TemplateDoesNotExist, engines

from django_slack_tools.slack_messages.message_templates import DjangoTemplate, PythonTemplate
from django_slack_tools.slack_messages.models import SlackMessagingPolicy

from .base import BaseTemplateLoader

if TYPE_CHECKING:
    from django.template.backends.base import BaseEngine


logger = logging.getLogger(__name__)


class DjangoTemplateLoader(BaseTemplateLoader):
    """Django filesystem-backed template loader."""

    def __init__(self, *, engine: BaseEngine | None = None) -> None:
        """Initialize template loader.

        Args:
            engine: Template engine to use. Defaults to Django engine.
        """
        self.engine = engines["django"] if engine is None else engine

    def load(self, key: str) -> DjangoTemplate | None:  # noqa: D102
        try:
            return DjangoTemplate(file=key, engine=self.engine)
        except TemplateDoesNotExist:
            logger.debug("Template not found: %s", key)
            return None


class DjangoPolicyTemplateLoader(BaseTemplateLoader):
    """Django database-backed template loader."""

    def load(self, key: str) -> PythonTemplate | DjangoTemplate | None:  # noqa: D102
        return self._get_template_from_policy(policy_or_code=key)

    def _get_template_from_policy(
        self,
        policy_or_code: SlackMessagingPolicy | str,
    ) -> PythonTemplate | DjangoTemplate | None:
        """Get template instance."""
        if isinstance(policy_or_code, str):
            try:
                policy = SlackMessagingPolicy.objects.get(code=policy_or_code)
            except SlackMessagingPolicy.DoesNotExist:
                logger.warning("Policy not found: %s", policy_or_code)
                return None
        else:
            policy = policy_or_code

        if policy.template_type == SlackMessagingPolicy.TemplateType.PYTHON:
            return PythonTemplate(policy.template)

        if policy.template_type == SlackMessagingPolicy.TemplateType.DJANGO:
            try:
                return DjangoTemplate(file=policy.template)
            except TemplateDoesNotExist:
                logger.debug("Template not found: %s", policy.template)
                return None

        if policy.template_type == SlackMessagingPolicy.TemplateType.DJANGO_INLINE:
            return DjangoTemplate(inline=policy.template)

        msg = f"Unsupported template type: {policy.template_type!r}"
        raise ValueError(msg)



from django_slack_tools.slack_messages.validators import body_validator, header_validator


def test_header_validator() -> None:
    header_validator({})
    header_validator(
        {
            "mrkdwn": None,
            "parse": None,
            "reply_broadcast": None,
            "thread_ts": None,
            "unfurl_links": None,
            "unfurl_media": None,
        },
    )


def test_body_validator() -> None:
    body_validator({"text": "some-text"})
    body_validator(
        {
            "attachments": None,
            "blocks": None,
            "text": "some-text",
            "icon_emoji": None,
            "icon_url": None,
            "metadata": None,
            "username": None,
        },
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from django_slack_tools.slack_messages.backends import DummyBackend
from django_slack_tools.slack_messages.message_templates import BaseTemplate, PythonTemplate
from django_slack_tools.slack_messages.middlewares import (
    BaseMiddleware,
)
from django_slack_tools.slack_messages.template_loaders import (
    BaseTemplateLoader,
)

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.request import MessageRequest
    from django_slack_tools.slack_messages.response import MessageResponse


class MockTemplate(BaseTemplate):
    def __init__(self, render: Callable[[dict[str, Any]], Any]) -> None:
        self._render = render

    def render(self, context: dict[str, Any]) -> Any:
        return self._render(context)


class MockTemplateLoader(BaseTemplateLoader):
    def __init__(self, template: Any = None, *, key: str | None = None) -> None:
        self.template = template or PythonTemplate({"text": "Hello, {name}!"})
        self.key = key

    def load(self, key: str) -> PythonTemplate | None:
        if self.key and key != self.key:
            return None

        return self.template


class MockMiddleware(BaseMiddleware):
    def __init__(
        self,
        *,
        process_request: Callable[[MessageRequest], MessageRequest | None] | None = None,
        process_response: Callable[[MessageResponse], MessageResponse | None] | None = None,
    ) -> None:
        self._process_request = process_request
        self._process_response = process_response

    def process_request(self, request: MessageRequest) -> MessageRequest | None:
        if self._process_request:
            return self._process_request(request)

        return super().process_request(request)

    def process_response(self, response: MessageResponse) -> MessageResponse | None:
        if self._process_response:
            return self._process_response(response)

        return super().process_response(response)


class MockBackend(DummyBackend):
    def __init__(self, *, should_error: bool = False) -> None:
        self.should_error = should_error

    def _send_message(self, *args: Any, **kwargs: Any) -> Any:
        if self.should_error:
            msg = "Some error occurred"
            raise Exception(msg)  # noqa: TRY002

        return super()._send_message(*args, **kwargs)



# Generated by Django 4.2.18 on 2025-03-13 14:03

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("slack_messages", "0005_alter_slackmessagingpolicy_template_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="slackmessage",
            name="id",
            field=models.CharField(
                default=uuid.uuid4,
                editable=False,
                max_length=255,
                primary_key=True,
                serialize=False,
                unique=True,
            ),
        ),
    ]



from __future__ import annotations

from datetime import datetime
from typing import Any

from factory import Factory, LazyAttribute, SubFactory, lazy_attribute  # type: ignore[attr-defined]

from django_slack_tools.slack_messages.request import MessageHeader, MessageRequest
from django_slack_tools.slack_messages.response import MessageResponse
from tests._factories import SlackResponseFactory


class MessageRequestFactory(Factory):
    channel = "some-channel"
    template_key = "some-template-key"
    context = {"some": "context"}  # noqa: RUF012
    header = LazyAttribute(lambda _: MessageHeader())

    class Meta:
        model = MessageRequest


class MessageResponseFactory(Factory):
    request = SubFactory(MessageRequestFactory)
    ok = True
    error = None
    data = {"some": "data"}  # noqa: RUF012
    ts = "some-ts"
    parent_ts = None

    class Meta:
        model = MessageResponse


class SlackMessageResponseFactory(SlackResponseFactory):
    """Response factory for `chat.postMessage` API method."""

    @lazy_attribute
    def data(self) -> dict[str, Any]:
        ts = str(datetime.now().timestamp())  # noqa: DTZ005
        return {
            "ok": True,
            "channel": "whatever-channel",
            "ts": ts,
            "message": {
                "bot_id": "<REDACTED>",
                "type": "message",
                "text": self.text,  # type: ignore[attr-defined]
                "user": "<REDACTED>",
                "ts": ts,
                "app_id": "<REDACTED>",
                "blocks": self.blocks,  # type: ignore[attr-defined]
                "team": "<REDACTED>",
                "bot_profile": {},
                "attachments": self.attachments,  # type: ignore[attr-defined]
            },
        }

    class Params:
        text = "Hello, World!"
        blocks = [  # noqa: RUF012
            {
                "type": "rich_text",
                "block_id": "FbU8",
                "elements": [
                    {
                        "type": "rich_text_section",
                        "elements": [
                            {
                                "type": "text",
                                "text": "Hello, World!",
                            },
                        ],
                    },
                ],
            },
        ]
        attachments = []  # type: ignore[var-annotated]  # noqa: RUF012


class SlackGetPermalinkResponseFactory(SlackResponseFactory):
    """Response factory for `chat.getPermalink` API method."""

    @lazy_attribute
    def data(self) -> dict[str, Any]:
        return {
            "ok": True,
            "channel": "whatever-channel",
            "permalink": self.permalink,  # type: ignore[attr-defined]
        }

    class Params:
        permalink = "https://ghostbusters.slack.com/archives/C1H9RESGA/p135854651500008"



"""Application settings."""
# flake8: noqa: UP007

from __future__ import annotations

from typing import Any, TypedDict, Union

from django.utils.module_loading import import_string
from typing_extensions import NotRequired

LazyInitSpec = TypedDict(
    "LazyInitSpec",
    {
        "class": str,
        "args": NotRequired[tuple],
        "kwargs": NotRequired[dict[str, Any]],
    },
)
LazyInitSupported = Union[LazyInitSpec, str]


def lazy_init(spec: LazyInitSupported) -> Any:
    """Initialize an object from a lazy init spec.

    If string is passed, it is treated as a class import path.
    Otherwise, it is expected to be a dictionary with keys:

    - `class`: class import path,
    - `args`: positional arguments,
    - `kwargs`: keyword arguments.
    """
    if isinstance(spec, str):
        spec = LazyInitSpec({"class": spec})

    class_ = import_string(spec["class"])
    args, kwargs = spec.get("args", ()), spec.get("kwargs", {})
    return class_(*args, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from django_slack_tools.slack_messages.admin import SlackMessageAdmin
from django_slack_tools.slack_messages.models import SlackMessage
from tests._helpers import ModelAdminTestBase
from tests.slack_messages.models._factories import SlackMessageFactory

if TYPE_CHECKING:
    from collections.abc import Iterable

    from django.test import Client
    from django.test.client import _MonkeyPatchedWSGIResponse


class TestSlackMessageAdmin(ModelAdminTestBase):
    admin_cls = SlackMessageAdmin
    model_cls = SlackMessage
    factory_cls = SlackMessageFactory

    pytestmark = pytest.mark.django_db()

    def _clone_messages(
        self,
        *,
        client: Client,
        ids: Iterable[SlackMessage],
    ) -> _MonkeyPatchedWSGIResponse:
        return client.post(
            self._reverse("changelist"),
            {
                "action": "_clone_messages",
                "_selected_action": ids,
            },
        )

    def _send_messages(
        self,
        *,
        client: Client,
        ids: Iterable[SlackMessage],
    ) -> _MonkeyPatchedWSGIResponse:
        return client.post(
            self._reverse("changelist"),
            {
                "action": "_send_messages",
                "_selected_action": ids,
            },
        )

    def test_changelist(self, admin_client: Client) -> None:
        # Test permalink field
        self.factory_cls.create(permalink="https://example.com")
        return super().test_changelist(admin_client)



# Generated by Django 4.2.17 on 2024-12-31 13:40
from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import migrations, models

if TYPE_CHECKING:
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor
    from django.db.migrations.state import StateApps


def _change_dict_to_python(apps: StateApps, schema_editor: BaseDatabaseSchemaEditor) -> None:
    db_alias = schema_editor.connection.alias
    model_type = apps.get_model("slack_messages", "SlackMessagingPolicy")
    model_type.objects.using(db_alias).filter(template_type="D").update(template_type="P")


def _revert_python_to_dict(apps: StateApps, schema_editor: BaseDatabaseSchemaEditor) -> None:
    db_alias = schema_editor.connection.alias
    model_type = apps.get_model("slack_messages", "SlackMessagingPolicy")
    model_type.objects.using(db_alias).filter(template_type="P").update(template_type="D")


class Migration(migrations.Migration):
    dependencies = [
        ("slack_messages", "0004_slackmessagingpolicy_template_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="slackmessagingpolicy",
            name="template_type",
            field=models.CharField(
                choices=[("P", "Python"), ("DJ", "Django"), ("DI", "Django Inline"), ("?", "Unknown")],
                default="P",
                help_text="Type of message template.",
                max_length=2,
                verbose_name="Template type",
            ),
        ),
        migrations.RunPython(
            code=_change_dict_to_python,
            reverse_code=_revert_python_to_dict,
        ),
    ]



# noqa: D100
# flake8: noqa: UP006, UP007, UP035
# ? Subscription syntax available since Python 3.10
from __future__ import annotations

from typing import Any, Optional

from pydantic import BaseModel

from django_slack_tools.slack_messages.request import MessageRequest  # noqa: TC001


class MessageResponse(BaseModel):
    """Response from a messaging backend."""

    request: Optional[MessageRequest] = None
    ok: bool
    error: Optional[Any] = None
    data: Any
    ts: Optional[str] = None
    parent_ts: Optional[str] = None



from .base import BaseTemplateLoader
from .django import DjangoPolicyTemplateLoader, DjangoTemplateLoader
from .errors import TemplateLoadError, TemplateNotFoundError

__all__ = (
    "BaseTemplateLoader",
    "DjangoPolicyTemplateLoader",
    "DjangoTemplateLoader",
    "TemplateLoadError",
    "TemplateNotFoundError",
)



# noqa: D100
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

_T = TypeVar("_T")


class BaseTemplate(ABC, Generic[_T]):
    """Base class for templates."""

    template: _T

    @abstractmethod
    def render(self, context: dict[str, Any]) -> Any:
        """Render the template with the given context."""



# noqa: D100
from __future__ import annotations

from typing import Any


def make_repr(obj: Any) -> str:
    """Make a repr string for an object."""
    args = ", ".join(f"{k}={v!r}" for k, v in obj.__dict__.items())
    return f"{obj.__class__.__name__}({args})"



from __future__ import annotations

import pytest

from django_slack_tools.slack_messages.backends import DummyBackend
from django_slack_tools.slack_messages.request import MessageBody, MessageHeader, MessageRequest
from django_slack_tools.slack_messages.response import MessageResponse


class TestDummyBackend:
    pytestmark = pytest.mark.django_db()

    @pytest.fixture
    def backend(self) -> DummyBackend:
        return DummyBackend()

    def test_deliver(self, backend: DummyBackend) -> None:
        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=MessageBody(text="Hello, World!"),
        )
        response = backend.deliver(request)

        assert isinstance(response, MessageResponse)
        assert response.request is request
        assert response.ok is True
        assert response.error is None
        assert response.data == {}
        assert response.ts is None
        assert response.parent_ts is None

    def test_deliver_request_body_required(self, backend: DummyBackend) -> None:
        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=None,
        )
        with pytest.raises(ValueError, match="Message body is required."):
            backend.deliver(request)



"""Slack messaging backends."""

from __future__ import annotations

import traceback
from abc import ABC, abstractmethod
from logging import getLogger
from typing import TYPE_CHECKING, Any, Optional, cast

from slack_sdk.errors import SlackApiError

from django_slack_tools.slack_messages.response import MessageResponse

if TYPE_CHECKING:
    from slack_sdk.web import SlackResponse

    from django_slack_tools.slack_messages.request import MessageBody, MessageHeader, MessageRequest


logger = getLogger(__name__)


class BaseBackend(ABC):
    """Abstract base class for messaging backends."""

    def deliver(self, request: MessageRequest) -> MessageResponse:
        """Deliver message request."""
        if request.body is None:
            msg = "Message body is required."
            raise ValueError(msg)

        try:
            response = self._send_message(
                channel=request.channel,
                header=request.header,
                body=request.body,
            )
            error = None
        except SlackApiError as err:
            response = err.response
            error = traceback.format_exc()

        ok = cast("bool", response.get("ok"))
        data: Any
        if ok:
            ts = cast("Optional[str]", response.get("ts", None))
            data = response.get("message", {})
            parent_ts = data.get("thread_ts", None)
        else:
            ts = None
            data = response.data
            parent_ts = None

        return MessageResponse(
            request=request,
            ok=ok,
            error=error,
            data=data,
            ts=ts,
            parent_ts=parent_ts,
        )

    @abstractmethod
    def _send_message(self, *, channel: str, header: MessageHeader, body: MessageBody) -> SlackResponse:
        """Internal implementation of actual 'send message' behavior."""



"""Celery utils."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from celery import shared_task
from celery.utils.log import get_task_logger
from django.utils import timezone

from django_slack_tools.slack_messages import shortcuts
from django_slack_tools.slack_messages.models import SlackMessage

if TYPE_CHECKING:
    from typing import Any

logger = get_task_logger(__name__)


@shared_task
def slack_message(
    to: str,
    *,
    messenger_name: str | None = None,
    template: str,
    header: dict[str, Any] | None = None,
    context: dict[str, Any] | None = None,
) -> str | None:
    """Celery task wrapper for `.shortcuts.slack_message`.

    Args:
        to: Recipient.
        messenger_name: Messenger name. If not set, default messenger is used.
        template: Message template key.
        header: Slack message control header.
        context: Context for rendering the template.

    Returns:
        ID of sent message if any, `None` otherwise.
    """
    response = shortcuts.slack_message(
        to,
        messenger_name=messenger_name,
        template=template,
        header=header,
        context=context,
    )
    return response.ts if response else None


@shared_task
def cleanup_old_messages(
    *,
    base_ts: str | None = None,
    threshold_seconds: int | None = 7 * 24 * 60 * 60,  # 7 days
) -> int:
    """Delete old messages created before given threshold.

    Args:
        threshold_seconds: Threshold seconds. Defaults to 7 days.
        base_ts: Base timestamp to calculate the threshold, in ISO format. If falsy, current timestamp will be used.

    Returns:
        Number of deleted messages.
    """
    dt = datetime.fromisoformat(base_ts) if base_ts else timezone.localtime()

    if threshold_seconds is None:
        logger.warning("Threshold seconds not provided, skipping cleanup.")
        return 0

    cleanup_threshold = dt - timedelta(seconds=threshold_seconds)
    logger.debug("Cleaning up messages older than %s.", cleanup_threshold)

    num_deleted, _ = SlackMessage.objects.filter(created__lt=cleanup_threshold).delete()
    logger.info("Deleted %d old messages.", num_deleted)

    return num_deleted



# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.request import MessageRequest
    from django_slack_tools.slack_messages.response import MessageResponse


class BaseMiddleware:
    """Base class for middleware components."""

    def process_request(self, request: MessageRequest) -> MessageRequest | None:  # pragma: no cover
        """Process the incoming requests.

        Args:
            request: Message request.

        Returns:
            MessageRequest objects or `None`.
        """
        return request

    def process_response(self, response: MessageResponse) -> MessageResponse | None:  # pragma: no cover
        """Processes a sequence of MessageResponse objects and returns the processed sequence.

        Args:
            response: Message response.

        Returns:
            MessageResponse objects or `None`.
        """
        return response



import pytest

from django_slack_tools.slack_messages.message_templates import DjangoTemplate, PythonTemplate
from django_slack_tools.slack_messages.models import SlackMessagingPolicy
from django_slack_tools.slack_messages.template_loaders import DjangoPolicyTemplateLoader, DjangoTemplateLoader
from tests.slack_messages.models._factories import SlackMessagingPolicyFactory


class TestDjangoTemplateLoader:
    def test_load(self) -> None:
        loader = DjangoTemplateLoader()
        template = loader.load("greet.xml")
        assert template
        assert isinstance(template, DjangoTemplate)
        assert template.template.template.name == "greet.xml"  # type: ignore[attr-defined] # Maybe false-positive error?

    def test_load_template_not_found(self) -> None:
        loader = DjangoTemplateLoader()
        template = loader.load("NOT_FOUND")
        assert template is None


class TestDjangoPolicyTemplateLoader:
    pytestmark = pytest.mark.django_db

    def test_load_python_template(self) -> None:
        policy = SlackMessagingPolicyFactory.create(
            template_type=SlackMessagingPolicy.TemplateType.PYTHON,
            template={
                "attachments": [
                    {
                        "blocks": [
                            {"text": {"text": "Hello, World!: {mentions}", "type": "mrkdwn"}, "type": "section"},
                        ],
                        "color": "#f2c744",
                    },
                ],
                "blocks": [{"text": {"text": "Hello, World!: {mentions}", "type": "mrkdwn"}, "type": "section"}],
            },
        )

        loader = DjangoPolicyTemplateLoader()
        template = loader.load(policy.code)

        assert template
        assert isinstance(template, PythonTemplate)
        assert template.template == {
            "attachments": [
                {
                    "blocks": [{"text": {"text": "Hello, World!: {mentions}", "type": "mrkdwn"}, "type": "section"}],
                    "color": "#f2c744",
                },
            ],
            "blocks": [{"text": {"text": "Hello, World!: {mentions}", "type": "mrkdwn"}, "type": "section"}],
        }

    def test_load_django_template_pass_policy_by_code(self) -> None:
        policy = SlackMessagingPolicyFactory.create(
            template_type=SlackMessagingPolicy.TemplateType.DJANGO,
            template="greet.xml",
        )

        loader = DjangoPolicyTemplateLoader()
        template = loader.load(policy.code)

        assert template
        assert isinstance(template, DjangoTemplate)
        assert template.template.template.name == "greet.xml"  # type: ignore[attr-defined] # Maybe false-positive error?

    def test_load_django_template_pass_policy_by_object(self) -> None:
        """Support passing policy object for convenience."""
        policy = SlackMessagingPolicyFactory.create(
            template_type=SlackMessagingPolicy.TemplateType.DJANGO,
            template="greet.xml",
        )

        loader = DjangoPolicyTemplateLoader()
        template = loader.load(policy)

        assert template
        assert isinstance(template, DjangoTemplate)
        assert template.template.template.name == "greet.xml"  # type: ignore[attr-defined] # Maybe false-positive error?

    def test_load_django_template_policy_not_found(self) -> None:
        loader = DjangoPolicyTemplateLoader()
        template = loader.load("NOT_FOUND")
        assert template is None

    def test_load_django_template_template_not_found(self) -> None:
        policy = SlackMessagingPolicyFactory.create(
            template_type=SlackMessagingPolicy.TemplateType.DJANGO,
            template="NOT_FOUND",
        )

        loader = DjangoPolicyTemplateLoader()
        template = loader.load(policy.code)

        assert template is None

    def test_load_django_inline_template(self) -> None:
        inline_template = """
<root>
    <block type="section">
        <text type="mrkdwn">
            {{ greet }}, {{ mentions }}
        </text>
    </block>
</root>
""".lstrip()
        policy = SlackMessagingPolicyFactory.create(
            template_type=SlackMessagingPolicy.TemplateType.DJANGO_INLINE,
            template=inline_template,
        )
        loader = DjangoPolicyTemplateLoader()
        template = loader.load(policy.code)

        assert template
        assert isinstance(template, DjangoTemplate)
        assert template.template.template.source == inline_template  # type: ignore[attr-defined] # Maybe false-positive error?

    def test_load_unknown_template_type(self) -> None:
        SlackMessagingPolicyFactory(
            code="TEST",
            template_type=SlackMessagingPolicy.TemplateType.UNKNOWN,
            template="unknown-template",
        )

        loader = DjangoPolicyTemplateLoader()
        with pytest.raises(ValueError, match="Unsupported template type: '?'"):
            loader.load("TEST")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.db.models import signals
from django.dispatch import receiver

from django_slack_tools.slack_messages.shortcuts import slack_message

from .models import Todo

if TYPE_CHECKING:
    from collections.abc import Iterable


@receiver(signals.post_save, sender=Todo)
def notify_for_create(
    sender: type[Todo],  # noqa: ARG001
    instance: Todo,
    created: bool,  # noqa: FBT001
    **kwargs: Any,  # noqa: ARG001
) -> None:
    """Send a Slack message if new To Do created or existing instance updated."""
    if created:
        slack_message("TODO-CREATE", context={"title": instance.title})


@receiver(signals.pre_save, sender=Todo)
def notify_for_update(
    sender: type[Todo],  # noqa: ARG001
    instance: Todo,
    update_fields: Iterable[str] | None = None,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> None:
    """Send a Slack message if new To Do created or existing instance updated."""
    try:
        old_instance = Todo.objects.get(pk=instance.pk)  # type: ignore[attr-defined]
    except Todo.DoesNotExist:
        return

    old_values = {
        "title": old_instance.title,
        "description": old_instance.description,
        "completed": old_instance.completed,
    }
    new_values = {
        "title": instance.title,
        "description": instance.description,
        "completed": instance.completed,
    }

    slack_message("TODO-UPDATE", context={"old_values": old_values, "new_values": new_values})



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast
from unittest import mock

import pytest

from django_slack_tools.slack_messages.messenger import Messenger
from django_slack_tools.slack_messages.middlewares import (
    DjangoDatabasePersister,
    DjangoDatabasePolicyHandler,
)
from django_slack_tools.slack_messages.request import MessageHeader, MessageRequest
from django_slack_tools.slack_messages.template_loaders import (
    DjangoPolicyTemplateLoader,
    DjangoTemplateLoader,
    TemplateNotFoundError,
)
from tests.slack_messages._factories import MessageRequestFactory

from ._helpers import MockBackend, MockMiddleware, MockTemplate, MockTemplateLoader

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.response import MessageResponse


class TestMessenger:
    def test_instance_creation(self) -> None:
        """Test various instance creation scenarios."""
        # Ideal case
        kwargs: dict[str, Any] = {
            "template_loaders": [
                DjangoTemplateLoader(),
                DjangoPolicyTemplateLoader(),
            ],
            "middlewares": [
                DjangoDatabasePersister(),
                DjangoDatabasePolicyHandler(messenger="default"),
            ],
            "messaging_backend": MockBackend(),
        }
        Messenger(**kwargs)

        # All template loaders should be inherited from `BaseTemplateLoader`
        with pytest.raises(
            TypeError,
            match=r"Expected inherited from <class '.+\.BaseTemplateLoader'>, got <class 'object'>",
        ):
            Messenger(**(kwargs | {"template_loaders": [DjangoTemplateLoader(), object()]}))

        # All middlewares should be inherited from `BaseMiddleware`
        with pytest.raises(
            TypeError,
            match=r"Expected inherited from <class '.+\.BaseMiddleware'>, got <class 'object'>",
        ):
            Messenger(**(kwargs | {"middlewares": [DjangoDatabasePersister(), object()]}))

        # Messaging backend should be inherited from `BaseBackend`
        with pytest.raises(
            TypeError,
            match=r"Expected inherited from <class '.+\.BaseBackend'>, got <class 'object'>",
        ):
            Messenger(**(kwargs | {"messaging_backend": object()}))

    def test_send(self) -> None:
        """Test `.send()` shortcut method, which wraps `.send_request()` for convenience."""
        messenger = Messenger(template_loaders=[], middlewares=[], messaging_backend=MockBackend())
        with mock.patch.object(messenger, "send_request") as send_request:
            messenger.send(to="channel", template="template", context={"key": "value"}, header={"thread_ts": "123"})
            # ? `send_request.assert_called_once_with` not applicable due to `id_`; unable to mock it
            assert send_request.call_count == 1
            assert cast("MessageRequest", send_request.call_args.kwargs["request"]).model_dump(
                exclude={"id_"},
            ) == MessageRequest(
                template_key="template",
                channel="channel",
                context={"key": "value"},
                header=MessageHeader(thread_ts="123"),
            ).model_dump(exclude={"id_"})

    def test_send_request(self) -> None:
        """Test sending a message request."""
        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware()],
            messaging_backend=MockBackend(),
        )
        response = messenger.send_request(request=MessageRequestFactory.create(context={"name": "Daniel"}))
        assert response
        assert response.model_dump() == {
            "data": {},
            "error": None,
            "ok": True,
            "parent_ts": None,
            "request": {
                "body": {
                    "attachments": None,
                    "blocks": None,
                    "icon_emoji": None,
                    "icon_url": None,
                    "metadata": None,
                    "text": "Hello, Daniel!",
                    "username": None,
                },
                "channel": "some-channel",
                "context": {"name": "Daniel"},
                "header": {
                    "mrkdwn": None,
                    "parse": None,
                    "reply_broadcast": None,
                    "thread_ts": None,
                    "unfurl_links": None,
                    "unfurl_media": None,
                },
                "id_": mock.ANY,
                "template_key": "some-template-key",
            },
            "ts": None,
        }

    def test_send_request_request_middleware_returned_none(self) -> None:
        """When request middleware returned `None`, the request should not be sent."""
        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware(process_request=lambda _: None)],
            messaging_backend=MockBackend(),
        )
        response = messenger.send_request(request=MessageRequestFactory.create())
        assert response is None

    def test_send_request_request_middleware_raised_error(self) -> None:
        """When request middleware raised an error, it should propagate the error."""

        def _throw_error(_: MessageRequest) -> None:
            msg = "Some error occurred"
            raise Exception(msg)  # noqa: TRY002

        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware(process_request=_throw_error)],
            messaging_backend=MockBackend(),
        )
        with pytest.raises(Exception, match="Some error occurred"):
            messenger.send_request(request=MessageRequestFactory.create())

    def test_send_request_template_key_is_not_set(self) -> None:
        """Template key is required to render the message."""
        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware()],
            messaging_backend=MockBackend(),
        )
        with pytest.raises(ValueError, match="Template key is required to render the message"):
            messenger.send_request(request=MessageRequestFactory.create(template_key=None))

    def test_send_request_template_load_not_found(self) -> None:
        """When template is not found, it should raise an error."""
        messenger = Messenger(
            template_loaders=[MockTemplateLoader(key="hacky-key")],
            middlewares=[MockMiddleware()],
            messaging_backend=MockBackend(),
        )
        with pytest.raises(TemplateNotFoundError, match="Template with key 'some-template-key' not found"):
            messenger.send_request(request=MessageRequestFactory.create(template_key="some-template-key"))

    def test_send_request_message_rendering_failed_missing_context_key(self) -> None:
        """Rendering fails if required context key is not provided."""
        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware()],
            messaging_backend=MockBackend(),
        )
        with pytest.raises(KeyError, match="name"):
            messenger.send_request(request=MessageRequestFactory.create(context={}))

    def test_send_request_message_rendering_failed_template_error_propagates_to_caller(self) -> None:
        """Error in rendering propagates to caller."""

        def _throw_error(_: dict[str, Any]) -> Any:
            msg = "Some error occurred"
            raise Exception(msg)  # noqa: TRY002

        messenger = Messenger(
            template_loaders=[MockTemplateLoader(MockTemplate(render=_throw_error))],
            middlewares=[MockMiddleware()],
            messaging_backend=MockBackend(),
        )
        with pytest.raises(Exception, match="Some error occurred"):
            messenger.send_request(request=MessageRequestFactory.create())

    def test_send_request_response_middleware_returned_none(self) -> None:
        """When response middleware returned `None`, it should return the `None`."""
        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware(process_response=lambda _: None)],
            messaging_backend=MockBackend(),
        )
        response = messenger.send_request(request=MessageRequestFactory.create(context={"name": "Daniel"}))
        assert response is None

    def test_send_request_response_middleware_raise_error(self) -> None:
        """When response middleware raised an error, it should propagate the error."""

        def _throw_error(_: MessageResponse) -> None:
            msg = "Some error occurred"
            raise Exception(msg)  # noqa: TRY002

        messenger = Messenger(
            template_loaders=[MockTemplateLoader()],
            middlewares=[MockMiddleware(process_response=_throw_error)],
            messaging_backend=MockBackend(),
        )
        with pytest.raises(Exception, match="Some error occurred"):
            messenger.send_request(request=MessageRequestFactory.create(context={"name": "Daniel"}))



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from django.core.exceptions import ImproperlyConfigured
from slack_bolt import App

from django_slack_tools.app_settings import AppSettings, get_settings_from_django
from django_slack_tools.slack_messages.messenger import Messenger

if TYPE_CHECKING:
    from pytest_django.fixtures import SettingsWrapper

    from django_slack_tools.app_settings import SettingsDict


# Config fixtures to run in parametrize (map of test id to config dictionary)
config_fixtures: dict[str, SettingsDict] = {
    "dummy backend": {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [],
                    "middlewares": [],
                    "messaging_backend": "django_slack_tools.slack_messages.backends.DummyBackend",
                },
            },
        },
    },
    "logging backend": {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [],
                    "middlewares": [],
                    "messaging_backend": "django_slack_tools.slack_messages.backends.LoggingBackend",
                },
            },
        },
    },
    "slack backend": {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [],
                    "middlewares": [],
                    "messaging_backend": {
                        "class": "django_slack_tools.slack_messages.backends.SlackBackend",
                        "kwargs": {
                            "slack_app": "testproj.config.slack_app.app",
                        },
                    },
                },
            },
        },
    },
    "slack redirect backend": {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [],
                    "middlewares": [],
                    "messaging_backend": {
                        "class": "django_slack_tools.slack_messages.backends.SlackRedirectBackend",
                        "kwargs": {
                            "slack_app": "testproj.config.slack_app.app",
                            "redirect_channel": "some-redirect-channel",
                        },
                    },
                },
            },
        },
    },
    "django db": {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [
                        "django_slack_tools.slack_messages.template_loaders.DjangoTemplateLoader",
                        "django_slack_tools.slack_messages.template_loaders.DjangoPolicyTemplateLoader",
                    ],
                    "middlewares": [
                        {
                            "class": "django_slack_tools.slack_messages.middlewares.DjangoDatabasePolicyHandler",
                            "kwargs": {
                                "messenger": "default",
                            },
                        },
                        "django_slack_tools.slack_messages.middlewares.DjangoDatabasePersister",
                    ],
                    "messaging_backend": {
                        "class": "django_slack_tools.slack_messages.backends.SlackBackend",
                        "kwargs": {
                            "slack_app": "testproj.config.slack_app.app",
                        },
                    },
                },
            },
        },
    },
}

# Save decorator for reuse, as not all test in suite requires this
settings_dict_parametrizer = pytest.mark.parametrize(
    argnames="settings_dict",
    argvalues=config_fixtures.values(),
    ids=config_fixtures.keys(),
)

# For import testing
not_slack_app = -1


class TestAppSettings:
    def _assert_app_settings(self, app_settings: AppSettings) -> None:
        assert isinstance(app_settings.slack_app, App)
        assert all(isinstance(messenger, Messenger) for messenger in app_settings.messengers.values())

    @settings_dict_parametrizer
    def test_dict_config(self, settings_dict: SettingsDict) -> None:
        app_settings = AppSettings.from_dict(settings_dict)
        self._assert_app_settings(app_settings)

    @settings_dict_parametrizer
    def test_django_config(self, settings: SettingsWrapper, settings_dict: SettingsDict) -> None:
        settings.DJANGO_SLACK_TOOLS = None
        with pytest.raises(ImproperlyConfigured, match=r"^Couldn't initialize app settings: .+"):
            get_settings_from_django()

        settings.DJANGO_SLACK_TOOLS = settings_dict
        app_settings = get_settings_from_django()
        self._assert_app_settings(app_settings)

    def test_bad_config_not_slack_app(self) -> None:
        with pytest.raises(
            ImproperlyConfigured,
            match="Expected <class 'slack_bolt.app.app.App'> instance, got <class 'int'>",
        ):
            AppSettings.from_dict(
                {
                    "slack_app": "tests.test_app_settings.not_slack_app",
                    "messengers": {},
                },
            )



from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest

from django_slack_tools.slack_messages.models import SlackMessage
from tests.slack_messages.models._factories import SlackMessageFactory

try:
    import celery  # noqa: F401
except ImportError:
    celery_installed = False
else:
    from django_slack_tools.slack_messages import tasks

    celery_installed = True

pytestmark = [
    pytest.mark.django_db,
    pytest.mark.skipif(not celery_installed, reason="Celery is not installed"),
]


class TestSlackMessage:
    def test_slack_message(self) -> None:
        with mock.patch("django_slack_tools.slack_messages.shortcuts.slack_message") as m:
            tasks.slack_message(
                "test",
                template="simple",
                header={"key": "value"},
                context={"greet": "Hello, world!"},
            )

        m.assert_called_once_with(
            "test",
            messenger_name=None,
            template="simple",
            header={"key": "value"},
            context={"greet": "Hello, world!"},
        )


class TestCleanupOldMessages:
    def test_cleanup_old_messages(self) -> None:
        # Arrange
        ts = datetime(2024, 10, 9, 3, 48, 22, tzinfo=timezone.utc)
        _should_deleted = [
            # Should be deleted
            SlackMessageFactory.create(created=ts - timedelta(minutes=6)),
            SlackMessageFactory.create(created=ts - timedelta(minutes=5, seconds=1)),
        ]
        should_remain = [
            # Should not be deleted
            SlackMessageFactory.create(created=ts - timedelta(minutes=5)),
            SlackMessageFactory.create(created=ts - timedelta(minutes=4, seconds=59)),
            SlackMessageFactory.create(created=ts - timedelta(minutes=4)),
        ]

        # Act
        num_deleted = tasks.cleanup_old_messages(base_ts=ts.isoformat(), threshold_seconds=5 * 60)  # 5 minutes

        # Assert
        assert num_deleted == 2
        assert sorted(SlackMessage.objects.values_list("id", flat=True)) == sorted(m.id for m in should_remain)

    def test_cleanup_old_messages_skip_if_threshold_is_none(self) -> None:
        # Arrange
        ts = datetime(2024, 10, 9, 3, 48, 22, tzinfo=timezone.utc)
        SlackMessageFactory(created=ts - timedelta(weeks=4))
        SlackMessageFactory(created=ts - timedelta(hours=1))
        SlackMessageFactory()

        # Act
        num_deleted = tasks.cleanup_old_messages(base_ts=ts.isoformat(), threshold_seconds=None)

        # Assert
        assert num_deleted == 0
        assert SlackMessage.objects.count() == 3



# flake8: noqa: D100, D103
from __future__ import annotations

import nox

nox.options.default_venv_backend = "uv"


# * Python and Django version is parametrized at the CI level for coverage tracking
@nox.session()
@nox.parametrize(
    # Parametrize tests with different combinations of extra dependencies
    "extras",
    [
        [],
        ["celery"],
    ],
)
def tests(
    session: nox.Session,
    *,
    extras: list[str],
) -> None:
    # Test dependencies are required and always installed
    extras.extend(["dev", "test"])

    # Run the tests via `uv`
    session.run_install("uv", "sync", "--quiet", *[f"--extra={extra}" for extra in extras])
    session.run("uv", "run", "pytest", "--cov-append")



# noqa: D100
# flake8: noqa: UP006, UP007, UP035
# #? Subscription syntax available since Python 3.10
from __future__ import annotations

import uuid
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator


class MessageRequest(BaseModel):
    """Message request object."""

    model_config = ConfigDict(extra="forbid")

    id_: str = Field(default_factory=lambda: str(uuid.uuid4()))
    channel: Any

    # Template key is optional to allow lazy initialization of the template key
    template_key: Optional[str]

    context: Dict[str, Any]
    header: MessageHeader

    # Also, the body is optional because it is rendered from the template
    body: Optional[MessageBody] = None


class MessageHeader(BaseModel):  # noqa: D101
    model_config = ConfigDict(extra="forbid")

    mrkdwn: Optional[str] = None
    parse: Optional[str] = None
    reply_broadcast: Optional[bool] = None
    thread_ts: Optional[str] = None
    unfurl_links: Optional[bool] = None
    unfurl_media: Optional[bool] = None

    @classmethod
    def from_any(cls, obj: MessageHeader | dict[str, Any] | None = None) -> MessageHeader:
        """Create instance from compatible types."""
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, dict):
            return cls.model_validate(obj)

        if obj is None:
            return cls()

        msg = f"Unsupported type {type(obj)}"
        raise TypeError(msg)


class MessageBody(BaseModel):  # noqa: D101
    model_config = ConfigDict(extra="forbid")

    attachments: Optional[List[dict]] = None

    # See more about blocks at https://api.slack.com/reference/block-kit/blocks
    blocks: Optional[list[dict]] = None

    text: Optional[str] = None
    icon_emoji: Optional[str] = None
    icon_url: Optional[str] = None
    metadata: Optional[dict] = None
    username: Optional[str] = None

    @model_validator(mode="after")
    def _check_at_least_one_field_is_set(self) -> MessageBody:
        if not any((self.attachments, self.blocks, self.text)):
            msg = "At least one of `attachments`, `blocks` and `text` must set"
            raise ValueError(msg)

        return self

    @classmethod
    def from_any(cls, obj: str | MessageBody | dict[str, Any]) -> MessageBody:
        """Create instance from compatible types."""
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, dict):
            return cls.model_validate(obj)

        if isinstance(obj, str):
            return cls(text=obj)

        msg = f"Unsupported type {type(obj)}"
        raise TypeError(msg)



# noqa: D100
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django_slack_tools.slack_messages.backends import BaseBackend
from django_slack_tools.slack_messages.middlewares import BaseMiddleware
from django_slack_tools.slack_messages.request import MessageBody, MessageHeader, MessageRequest
from django_slack_tools.slack_messages.template_loaders import BaseTemplateLoader, TemplateNotFoundError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from django_slack_tools.slack_messages.message_templates import BaseTemplate
    from django_slack_tools.slack_messages.response import MessageResponse

logger = logging.getLogger(__name__)


class Messenger:
    """Messenger class that sends message using templates and middlewares.

    Components evaluated in order:

    1. Request processing middlewares
    2. Load template by key and render message in-place
    4. Send message
    5. Response processing middlewares (in reverse order)
    """

    def __init__(
        self,
        *,
        template_loaders: Sequence[BaseTemplateLoader],
        middlewares: Sequence[BaseMiddleware],
        messaging_backend: BaseBackend,
    ) -> None:
        """Initialize the Messenger.

        Args:
            template_loaders: A sequence of template loaders.
                It is tried in order to load the template and the first one that returns a template is used.
            middlewares: A sequence of middlewares.
                Middlewares are applied in the order they are provided for request, and in reverse order for response.
            messaging_backend: The messaging backend to be used.
        """
        # Validate the template loaders
        for tl in template_loaders:
            if not isinstance(tl, BaseTemplateLoader):
                msg = f"Expected inherited from {BaseTemplateLoader!s}, got {type(tl)}"
                raise TypeError(msg)

        self.template_loaders = template_loaders

        # Validate the middlewares
        for mw in middlewares:
            if not isinstance(mw, BaseMiddleware):
                msg = f"Expected inherited from {BaseMiddleware!s}, got {type(mw)}"
                raise TypeError(msg)

        self.middlewares = middlewares

        # Validate the messaging backend
        if not isinstance(messaging_backend, BaseBackend):
            msg = f"Expected inherited from {BaseBackend!s}, got {type(messaging_backend)}"
            raise TypeError(msg)

        self.messaging_backend = messaging_backend

    def send(
        self,
        to: str,
        *,
        template: str | None = None,
        context: dict[str, str],
        header: MessageHeader | dict[str, Any] | None = None,
    ) -> MessageResponse | None:
        """Simplified shortcut for `.send_request()`."""
        header = MessageHeader.model_validate(header or {})
        request = MessageRequest(template_key=template, channel=to, context=context, header=header)
        return self.send_request(request=request)

    def send_request(self, request: MessageRequest) -> MessageResponse | None:
        """Sends a message request and processes the response."""
        logger.info("Sending request: %s", request)
        _request = self._process_request(request)
        if _request is None:
            return None

        self._render_message(_request)
        response = self._deliver_message(_request)
        _response = self._process_response(response)
        if _response is None:
            return None

        logger.info("Response: %s", _response)
        return response

    def _process_request(self, request: MessageRequest) -> MessageRequest | None:
        """Processes the request with middlewares in forward order."""
        for middleware in self.middlewares:
            logger.debug("Processing request (%s) with middleware %s", request, middleware)
            new_request = middleware.process_request(request)
            if new_request is None:
                logger.warning("Middleware %s returned `None`, skipping remaining middlewares", middleware)
                return None

            request = new_request

        logger.debug("Request after processing: %s", request)
        return request

    def _render_message(self, request: MessageRequest) -> None:
        """Updates the request with rendered message, in-place."""
        if request.body is not None:
            logger.debug("Request already has a body, skipping rendering")
            return

        if request.template_key is None:
            msg = "Template key is required to render the message"
            raise ValueError(msg)

        template = self._get_template(request.template_key)
        logger.debug("Rendering request %s with template: %s", request, template)
        rendered = template.render(request.context)
        request.body = MessageBody.model_validate(rendered)

    def _get_template(self, key: str) -> BaseTemplate:
        """Loads the template by key."""
        for loader in self.template_loaders:
            template = loader.load(key)
            if template is not None:
                return template

        msg = f"Template with key '{key}' not found"
        raise TemplateNotFoundError(msg)

    def _deliver_message(self, request: MessageRequest) -> MessageResponse:
        """Invoke the messaging backend to deliver the message."""
        logger.debug("Delivering message request: %s", request)
        response = self.messaging_backend.deliver(request)
        logger.debug("Response after delivery: %s", response)
        return response

    def _process_response(self, response: MessageResponse) -> MessageResponse | None:
        """Processes the response with middlewares in reverse order."""
        for middleware in reversed(self.middlewares):
            logger.debug("Processing response (%s) with middleware: %s", response, middleware)
            new_response = middleware.process_response(response)
            if new_response is None:
                logger.warning("Middleware %s returned `None`, skipping remaining middlewares", middleware)
                return None

            response = new_response

        logger.debug("Response after processing: %s", response)
        return response



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from slack_bolt import App

from django_slack_tools.slack_messages.backends import SlackBackend, SlackRedirectBackend
from django_slack_tools.slack_messages.request import MessageBody, MessageHeader, MessageRequest
from django_slack_tools.slack_messages.response import MessageResponse
from tests._factories import SlackApiErrorFactory
from tests.slack_messages._factories import SlackMessageResponseFactory

if TYPE_CHECKING:
    from unittest.mock import Mock


# Values for import testing
_slack_app = App(
    token="stupid-sandwich",  # noqa: S106
    signing_secret="peanut-butter",  # noqa: S106
    token_verification_enabled=False,
)
_not_slack_app = -1


class TestSlackBackend:
    pytestmark = pytest.mark.django_db()

    @pytest.fixture
    def backend(self, slack_app: App) -> SlackBackend:
        return SlackBackend(slack_app=slack_app)

    def test_instance_creation(self) -> None:
        assert SlackBackend(slack_app="tests.slack_messages.backends.test_slack._slack_app")
        assert SlackBackend(slack_app=_slack_app)
        with pytest.raises(
            TypeError,
            match="Expected <class 'slack_bolt.app.app.App'> instance, got <class 'int'>",
        ):
            SlackBackend(slack_app="tests.slack_messages.backends.test_slack._not_slack_app")

    def test_deliver(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        """Test sending message."""
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=MessageBody(text="Hello, World!"),
        )
        response = backend.deliver(request)

        assert isinstance(response, MessageResponse)
        assert response.request is request
        assert response.ok is True
        assert response.error is None
        assert response.data
        assert response.ts
        assert response.parent_ts is None

    def test_deliver_request_body_required(self, backend: SlackBackend) -> None:
        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=None,
        )
        with pytest.raises(ValueError, match="Message body is required."):
            backend.deliver(request)

    def test_deliver_remote_api_error(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        """Test sending message with error."""
        mock_slack_client.chat_postMessage.side_effect = SlackApiErrorFactory()

        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=MessageBody(text="Hello, World!"),
        )
        response = backend.deliver(request)

        assert isinstance(response, MessageResponse)
        assert response.request is request
        assert response.ok is False
        assert response.error
        assert response.data == {
            "ok": False,
        }
        assert response.ts is None
        assert response.parent_ts is None


class TestSlackRedirectBackend:
    pytestmark = pytest.mark.django_db()

    @pytest.fixture
    def backend(self, slack_app: App) -> SlackRedirectBackend:
        return SlackRedirectBackend(slack_app=slack_app, redirect_channel="test-redirect-channel")

    def test_deliver(self, backend: SlackRedirectBackend, mock_slack_client: Mock) -> None:
        """Test sending message."""
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=MessageBody(text="Hello, World!"),
        )
        response = backend.deliver(request)

        assert isinstance(response, MessageResponse)
        assert response.request is request
        assert response.ok is True
        assert response.error is None
        assert response.data
        assert response.ts
        assert response.parent_ts is None
        mock_slack_client.chat_postMessage.assert_called_once_with(
            channel="test-redirect-channel",
            mrkdwn=None,
            parse=None,
            reply_broadcast=None,
            thread_ts=None,
            unfurl_links=None,
            unfurl_media=None,
            attachments=[
                {
                    "color": "#eb4034",
                    "text": ":warning:  This message was originally sent to channel *test-channel* but redirected here.",  # noqa: E501
                },
            ],
            blocks=None,
            text="Hello, World!",
            icon_emoji=None,
            icon_url=None,
            metadata=None,
            username=None,
        )

    def test_deliver_request_body_required(self, backend: SlackRedirectBackend) -> None:
        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=None,
        )
        with pytest.raises(ValueError, match="Message body is required."):
            backend.deliver(request)

    def test_deliver_remote_api_error(self, backend: SlackRedirectBackend, mock_slack_client: Mock) -> None:
        """Test sending message with error."""
        mock_slack_client.chat_postMessage.side_effect = SlackApiErrorFactory()

        request = MessageRequest(
            channel="test-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=MessageBody(text="Hello, World!"),
        )
        response = backend.deliver(request)

        assert isinstance(response, MessageResponse)
        assert response.request is request
        assert response.ok is False
        assert response.error
        assert response.data == {
            "ok": False,
        }
        assert response.ts is None
        assert response.parent_ts is None

    def test_deliver_disable_inform_redirect(self, backend: SlackRedirectBackend, mock_slack_client: Mock) -> None:
        backend.inform_redirect = False
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

        request = MessageRequest(
            channel="whatever-this-channel",
            template_key="__any__",
            context={},
            header=MessageHeader(),
            body=MessageBody(text="Hello, World!"),
        )
        response = backend.deliver(request)

        assert isinstance(response, MessageResponse)
        mock_slack_client.chat_postMessage.assert_called_once_with(
            channel="test-redirect-channel",
            mrkdwn=None,
            parse=None,
            reply_broadcast=None,
            thread_ts=None,
            unfurl_links=None,
            unfurl_media=None,
            attachments=None,
            blocks=None,
            text="Hello, World!",
            icon_emoji=None,
            icon_url=None,
            metadata=None,
            username=None,
        )



# noqa: D100
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.message_templates import BaseTemplate


class BaseTemplateLoader(ABC):
    """Base class for template loaders."""

    @abstractmethod
    def load(self, key: str) -> BaseTemplate | None:
        """Load a template by key."""



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from django_slack_tools.slack_messages.response import MessageResponse
from django_slack_tools.slack_messages.shortcuts import slack_message

from ._factories import SlackMessageResponseFactory

if TYPE_CHECKING:
    from unittest.mock import Mock

    from django_slack_tools.app_settings import SettingsDict

pytestmark = [
    pytest.mark.django_db,
    pytest.mark.usefixtures("override_app_settings"),
]


@pytest.fixture(scope="session")
def app_settings() -> SettingsDict:
    return {
        "slack_app": "testproj.config.slack_app.app",
        "messengers": {
            "default": {
                "class": "django_slack_tools.slack_messages.messenger.Messenger",
                "kwargs": {
                    "template_loaders": [
                        "django_slack_tools.slack_messages.template_loaders.DjangoTemplateLoader",
                        "django_slack_tools.slack_messages.template_loaders.DjangoPolicyTemplateLoader",
                    ],
                    "middlewares": [
                        "django_slack_tools.slack_messages.middlewares.DjangoDatabasePersister",
                    ],
                    "messaging_backend": {
                        "class": "django_slack_tools.slack_messages.backends.SlackBackend",
                        "kwargs": {
                            "slack_app": "testproj.config.slack_app.app",
                        },
                    },
                },
            },
        },
    }


def test_slack_message_simplest_message(mock_slack_client: Mock) -> None:
    # Arrange
    mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

    # Act
    response = slack_message("whatever-channel", message="Hello, World!")

    # Assert
    assert isinstance(response, MessageResponse)
    assert (request := response.request)
    assert request.model_dump() == {
        "body": {
            "attachments": None,
            "blocks": None,
            "icon_emoji": None,
            "icon_url": None,
            "metadata": None,
            "text": "Hello, World!",
            "username": None,
        },
        "channel": "whatever-channel",
        "context": {},
        "header": {
            "mrkdwn": None,
            "parse": None,
            "reply_broadcast": None,
            "thread_ts": None,
            "unfurl_links": None,
            "unfurl_media": None,
        },
        "id_": mock.ANY,
        "template_key": None,
    }

    assert response.error is None
    assert response.data
    assert response.ts
    assert response.parent_ts is None


def test_slack_message_with_template(mock_slack_client: Mock) -> None:
    # Arrange
    mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

    # Act
    response = slack_message("whatever-channel", template="greet.xml", context={"greet": "Hello, World!"})

    # Assert
    assert isinstance(response, MessageResponse)
    assert (request := response.request)
    assert request.model_dump() == {
        "body": {
            "attachments": None,
            "blocks": [{"text": {"text": "Hello, World!,", "type": "mrkdwn"}, "type": "section"}],
            "icon_emoji": None,
            "icon_url": None,
            "metadata": None,
            "text": None,
            "username": None,
        },
        "channel": "whatever-channel",
        "context": {
            "greet": "Hello, World!",
        },
        "header": {
            "mrkdwn": None,
            "parse": None,
            "reply_broadcast": None,
            "thread_ts": None,
            "unfurl_links": None,
            "unfurl_media": None,
        },
        "id_": mock.ANY,
        "template_key": "greet.xml",
    }
    assert response.error is None
    assert response.data
    assert response.ts
    assert response.parent_ts is None


def test_slack_message_mutually_exclusive_arguments(mock_slack_client: Mock) -> None:
    # Arrange
    mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

    # Act & Assert
    with pytest.raises(ValueError, match="Either `template` or `message` must be set, but not both."):
        slack_message(  # type: ignore[call-overload]
            "whatever-channel",
            message="Hello, World!",
            template="greet.xml",
            context={"greet": "Hello, World!"},
        )

    mock_slack_client.chat_postMessage.assert_not_called()



# noqa: D100
from __future__ import annotations

import logging
from typing import Any, TypeVar

from .base import BaseTemplate

_PyObj = TypeVar("_PyObj", dict, list, str)

logger = logging.getLogger(__name__)


class PythonTemplate(BaseTemplate[_PyObj]):
    """Template that renders a dictionary."""

    def __init__(self, template: _PyObj) -> None:
        """Initialize the template."""
        self.template = template

    def render(self, context: dict[str, Any]) -> _PyObj:  # noqa: D102
        logger.debug("Rendering template %r with context %r", self.template, context)
        result = _format_obj(self.template, context=context)
        logger.debug("Rendered template %r to %r", self.template, result)
        return result


def _format_obj(obj: _PyObj, *, context: dict[str, Any]) -> _PyObj:
    if isinstance(obj, dict):
        return {key: _format_obj(value, context=context) for key, value in obj.items()}

    if isinstance(obj, list):
        return [_format_obj(item, context=context) for item in obj]

    if isinstance(obj, str):
        return obj.format_map(context)

    return obj



"""Slack backends actually interact with Slack API to do something."""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any

from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from slack_bolt import App

from .base import BaseBackend

if TYPE_CHECKING:
    from slack_sdk.web import SlackResponse

    from django_slack_tools.slack_messages.request import MessageBody, MessageHeader


logger = getLogger(__name__)


class SlackBackend(BaseBackend):
    """Backend actually sending the messages."""

    def __init__(self, *, slack_app: App | str) -> None:
        """Initialize backend.

        Args:
            slack_app: Slack app instance or import string.
        """
        if isinstance(slack_app, str):
            slack_app = import_string(slack_app)

        if not isinstance(slack_app, App):
            msg = f"Expected {App!s} instance, got {type(slack_app)}"
            raise TypeError(msg)

        self._slack_app = slack_app

    def _send_message(self, *, channel: str, header: MessageHeader, body: MessageBody) -> SlackResponse:
        return self._slack_app.client.chat_postMessage(
            channel=channel,
            **header.model_dump(),
            **body.model_dump(),
        )


class SlackRedirectBackend(SlackBackend):
    """Inherited Slack backend with redirection to specific channels."""

    def __init__(self, *, slack_app: App | str, redirect_channel: str, inform_redirect: bool = True) -> None:
        """Initialize backend.

        Args:
            slack_app: Slack app instance or import string.
            redirect_channel: Slack channel to redirect.
            inform_redirect: Whether to append an attachment informing that the message has been redirected.
                Defaults to `True`.
        """
        self.redirect_channel = redirect_channel
        self.inform_redirect = inform_redirect

        super().__init__(slack_app=slack_app)

    def _send_message(self, *, channel: str, header: MessageHeader, body: MessageBody) -> SlackResponse:
        if self.inform_redirect:
            attachments = body.attachments or []
            attachments.append(
                self._make_inform_attachment(original_channel=channel),
            )
            body.attachments = attachments

        return self._slack_app.client.chat_postMessage(
            channel=self.redirect_channel,
            **header.model_dump(),
            **body.model_dump(),
        )

    def _make_inform_attachment(self, *, original_channel: str) -> dict[str, Any]:
        msg_redirect_inform = _(
            ":warning:  This message was originally sent to channel *{channel}* but redirected here.",
        )

        return {
            "color": "#eb4034",
            "text": msg_redirect_inform.format(channel=original_channel),
        }



"""Handy APIs for sending Slack messages."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, overload

from django_slack_tools.app_settings import get_messenger
from django_slack_tools.slack_messages.request import MessageBody, MessageHeader, MessageRequest

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.response import MessageResponse

logger = logging.getLogger(__name__)


@overload
def slack_message(
    to: str,
    *,
    messenger_name: str | None = None,
    header: MessageHeader | dict[str, Any] | None = None,
    message: str,
) -> MessageResponse | None: ...  # pragma: no cover


@overload
def slack_message(
    to: str,
    *,
    messenger_name: str | None = None,
    header: MessageHeader | dict[str, Any] | None = None,
    template: str | None = None,
    context: dict[str, Any] | None = None,
) -> MessageResponse | None: ...  # pragma: no cover


def slack_message(  # noqa: PLR0913
    to: str,
    *,
    messenger_name: str | None = None,
    header: MessageHeader | dict[str, Any] | None = None,
    template: str | None = None,
    context: dict[str, Any] | None = None,
    message: str | None = None,
) -> MessageResponse | None:
    """Shortcut for sending a Slack message.

    Args:
        to: Recipient.
        messenger_name: Messenger name. If not set, default messenger is used.
        header: Slack message control header.
        template: Message template key. Cannot be used with `message`.
        context: Context for rendering the template. Only used with `template`.
        message: Simple message text. Cannot be used with `template`.

    Returns:
        Sent message instance or `None`.
    """
    if (template and message) or (not template and not message):
        msg = "Either `template` or `message` must be set, but not both."
        raise ValueError(msg)

    messenger = get_messenger(messenger_name)
    header = MessageHeader.from_any(header)

    if message:
        request = MessageRequest(
            channel=to,
            header=header,
            body=MessageBody(text=message),
            template_key=None,
            context={},
        )
        return messenger.send_request(request)

    context = context or {}
    return messenger.send(to, header=header, template=template, context=context)
